/*
 * Copyright (c) Infinitegra Inc, All rights reserved.
 */

package infinitegra.owlift.sample.owliftsampleapp;

import android.content.Context;
import android.graphics.PixelFormat;
import android.graphics.Rect;
import android.opengl.GLSurfaceView;
import android.opengl.GLU;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;

public class GLView extends GLSurfaceView implements
	GLSurfaceView.Renderer
{
	private boolean textureGen, sizeSet, surfaceCreated;
	private int textureId;
	private int frW, frH, txW, txH, drW, drH, drX, drY;
	private ByteBuffer image;
	private FrameRate frameRate = new FrameRate();
	private int[] intAry1 = new int[1];
	private FloatBuffer squareVerticesBuf, textureCoordsBuf;

	public GLView(Context c)
	{
		super(c);

		setRenderer(this);
		setRenderMode(RENDERMODE_WHEN_DIRTY);

		getHolder().setFormat(PixelFormat.TRANSLUCENT);
	}

	public FrameRate getFrameRate()
	{
		return frameRate;
	}

	private void setVertices(float[] textureCoords, float[] squareVertices)
	{
		ByteBuffer byteBuffer = ByteBuffer
			.allocateDirect(squareVertices.length * 4);
		byteBuffer.order(ByteOrder.nativeOrder());
		squareVerticesBuf = byteBuffer.asFloatBuffer();
		squareVerticesBuf.put(squareVertices);
		squareVerticesBuf.position(0);

		byteBuffer = ByteBuffer.allocateDirect(textureCoords.length * 4);
		byteBuffer.order(ByteOrder.nativeOrder());
		textureCoordsBuf = byteBuffer.asFloatBuffer();
		textureCoordsBuf.put(textureCoords);
		textureCoordsBuf.position(0);
	}

	private void calcDrawRect(Rect r)
	{
		float tw = (float)frW / (float)txW;
		float th = (float)frH / (float)txH;

		int rw = r.right - r.left;
		int rh = r.bottom - r.top;
		float sw, sh;

		if (rw > rh) {
			if ((frH != 0) && frW * 9 / frH == 16) {
				sw = 1.35f;
				sh = 0.76f;
			} else {
				sw = 1.0f;
				sh = (float)frH / (float)frW;
			}
		} else {
			sw = rw / rh;
			sh = sw * ((float)frH / (float)frW) * 0.9f;
		}

		if (rw > rh &&
			(float)rw / (float)rh > (float)frW / (float)frH) {
			drH = rh;
			drW = (int)((float)frW / frH * drH);
			drX = (rw - drW) / 2;
			drY = 0;
		} else {
			drW = rw;
			drH = (int)((float)frH / frW * drW);
			drX = 0;
			drY = (rh - drH) / 2;
		}

		float[] textureCoords = new float[] {
			// top left (V2)
			0.0f, th,
			// bottom left (V1)
			0.0f, 0.0f,
			// top right (V4)
			tw, th,
			// bottom right (V3)
			tw, 0.0f
		};

		float[] squareVertices = new float[] {
			// V1 - bottom left
			-sw, -sh, 0.0f,
			// V2 - top left
			-sw, sh, 0.0f,
			// V3 - bottom right
			sw, -sh, 0.0f,
			// V4 - top right
			sw, sh, 0.0f
		};

		setVertices(textureCoords, squareVertices);
	}

	public boolean getReady()
	{
		return sizeSet && surfaceCreated;
	}

	public void setImageSize(int width, int height, int texWidth, int texHeight)
	{
		this.frW = width;
		this.frH = height;
		this.txW = texWidth;
		this.txH = texHeight;

		Rect r = getHolder().getSurfaceFrame();
		if (r.right > 0)
			calcDrawRect(r);
		sizeSet = true;
	}

	@Override
	public void onSurfaceChanged(GL10 gl, int width, int height)
	{
		Rect r = getHolder().getSurfaceFrame();
		if (r.right > 0)
			calcDrawRect(r);

		surfaceChanged(gl, width, height);
	}

	@Override
	public void onSurfaceCreated(GL10 gl, EGLConfig config)
	{
		if (frW > 0)
			calcDrawRect(getHolder().getSurfaceFrame());

		createTexture(gl);

		gl.glEnable(GL10.GL_TEXTURE_2D);
		gl.glClearColor(0.0f, 0.0f, 0.0f, 0.5f);
		gl.glClearDepthf(1.0f);

		surfaceCreated = true;
	}

	private void createTexture(GL10 gl)
	{
		if (!textureGen) {
			intAry1[0] = textureId;
			gl.glDeleteTextures(1, intAry1, 0);
		}

		gl.glGenTextures(1, intAry1, 0);
		textureId = intAry1[0];
		gl.glBindTexture(GL10.GL_TEXTURE_2D, textureId);
		gl.glTexParameterf(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MIN_FILTER,
			GL10.GL_NEAREST);
		gl.glTexParameterf(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MAG_FILTER,
			GL10.GL_LINEAR);
		gl.glTexParameterf(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_WRAP_S,
			GL10.GL_CLAMP_TO_EDGE);
		gl.glTexParameterf(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_WRAP_T,
			GL10.GL_CLAMP_TO_EDGE);
		gl.glTexEnvf(GL10.GL_TEXTURE_ENV, GL10.GL_TEXTURE_ENV_MODE,
			GL10.GL_REPLACE);

		textureGen = true;
	}

	@Override
	public void onDrawFrame(GL10 gl)
	{
		drawFrame(gl);
	}

	public void convertFrameCoordinates(float x, float y, int w, int h,
		int[] coordinates)
	{
		if (coordinates == null || coordinates.length < 2) {
			return;
		}
		float w2d = x - drX;
		float h2d = y - drY;
		w2d = (float)Math.sqrt((w2d * w2d) + (2 * 2));
		h2d = (float)Math.sqrt((h2d * h2d) + (2 * 2));

		coordinates[0] = (int)w2d * w / drW;
		coordinates[1] = (int)h2d * h / drH;
	}

	public void render(byte[] image)
	{
		this.image = ByteBuffer.wrap(image);
		requestRender();
	}

	private void surfaceChanged(GL10 gl, int width, int height)
	{
		if (height == 0)
			height = 1;

		gl.glViewport(0, 0, width, height);
		gl.glMatrixMode(GL10.GL_PROJECTION);
		gl.glLoadIdentity();

		GLU.gluPerspective(gl, 45.0f, (float)width / (float)height, 1f,
			100.0f);

		gl.glMatrixMode(GL10.GL_MODELVIEW);
		gl.glLoadIdentity();
		GLU.gluLookAt(gl, 0, 0, 5.1f, 0, 0, 0, 0, 1, 0);
		gl.glNormal3f(0, 0, 1);
	}

	private void drawFrame(GL10 gl)
	{
		if (!sizeSet) {
			return;
		}
		frameRate.record();

		gl.glClear(GL10.GL_COLOR_BUFFER_BIT | GL10.GL_DEPTH_BUFFER_BIT);

		if (image == null)
			return;

		gl.glLoadIdentity();
		gl.glTranslatef(0.0f, 0.0f, -2.0f);
		gl.glBindTexture(GL10.GL_TEXTURE_2D, textureId);
		gl.glTexImage2D(GL10.GL_TEXTURE_2D, 0, GL10.GL_RGB,
			txW, txH, 0, GL10.GL_RGB,
			GL10.GL_UNSIGNED_SHORT_5_6_5, image);
		gl.glEnableClientState(GL10.GL_VERTEX_ARRAY);
		gl.glEnableClientState(GL10.GL_TEXTURE_COORD_ARRAY);
		gl.glFrontFace(GL10.GL_CW);
		gl.glVertexPointer(3, GL10.GL_FLOAT, 0, squareVerticesBuf);
		gl.glTexCoordPointer(2, GL10.GL_FLOAT, 0, textureCoordsBuf);
		gl.glDrawArrays(GL10.GL_TRIANGLE_STRIP, 0, 4);
		gl.glDisableClientState(GL10.GL_VERTEX_ARRAY);
		gl.glDisableClientState(GL10.GL_TEXTURE_COORD_ARRAY);
	}

	@Override
	public boolean performClick()
	{
		return super.performClick();
	}
}
