/*
 * Copyright (c) Infinitegra Inc, All rights reserved.
 */

package infinitegra.owlift.sample.owliftsampleapp;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.Dialog;
import android.app.DialogFragment;
import android.content.Context;
import android.content.SharedPreferences;
import android.content.SharedPreferences.Editor;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.os.AsyncTask;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.util.Log;
import android.view.Gravity;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnTouchListener;
import android.view.ViewGroup;
import android.view.Window;
import android.view.WindowManager;
import android.widget.CheckBox;
import android.widget.CompoundButton;
import android.widget.FrameLayout;
import android.widget.LinearLayout;
import android.widget.NumberPicker;
import android.widget.RadioGroup;
import android.widget.SeekBar;
import android.widget.Switch;
import android.widget.TextView;
import android.widget.Toast;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigDecimal;
import java.nio.ByteBuffer;
import java.util.Calendar;
import java.util.Locale;

import infinitegra.owlift.DecodeFunc;
import infinitegra.owlift.OwliftApp;
import infinitegra.owlift.OwliftErrorId;
import infinitegra.owlift.OwliftException;
import infinitegra.owlift.OwliftExceptionCode;
import infinitegra.owlift.OwliftMediaRecorder;
import infinitegra.owlift.OwliftMic;
import infinitegra.owlift.UsbOwliftDev;
import infinitegra.owlift.UsbOwliftFunc;
import infinitegra.owlift.UsbOwliftHost;
import infinitegra.owlift.UvcOwliftFunc;
import infinitegra.owlift.UvcOwliftFunc.UncompEncodingType;

public class MainActivity extends Activity implements OwliftDialogCallbacks,
	OnTouchListener
{
	// 拡大のサイズ。
	public static final int OWLIFT_MAGNIFY_SIZE = 3;

	public static final int OWLIFT_RANGE_MIN_TEMP_DEFAULT =
		(int)((273.15 + 16.0) * 100);
	public static final int OWLIFT_RANGE_MAX_TEMP_DEFAULT =
		(int)((273.15 + 48.0) * 100);

	public static final int OWLIFT_TEMP_MIN = 27315;
	public static final int OWLIFT_TEMP_MAX = 35315;

	private static final String DEFAULT_RECORD_DIR =
		Environment.getExternalStorageDirectory().getPath() + "/DCIM/owlift";

	public static final double OWLIFT_DEFAULT_EMISSIVITY = 0.95;

	private static boolean isCreated = false;
	private static boolean isPaused = false;

	private static final String TAG = "OwliftPlayer";
	private static final String PREF_NAME = "OWLIFT";
	private static final String PREF_USE_HOSTMIC = "useHostMic";

	// pref owlift
	private static final String PREF_OWLIFT_MIN_TEMP = "owliftmintemp";
	private static final String PREF_OWLIFT_MAX_TEMP = "owliftmaxtemp";
	private static final String PREF_OWLIFT_TEMP_UNIT = "owlifttempunit";
	private static final String PREF_OWLIFT_COLOR_TYPE = "owliftcolortype";
	private static final String PREF_OWLIFT_TEMP_MARKER_HIGH =
		"owlifttempmarkhigh";
	private static final String PREF_OWLIFT_TEMP_MARKER_LOW =
		"owlifttempmarklow";
	private static final String PREF_OWLFIT_DISPlAY_DATE = "owliftdispdate";
	private static final String PREF_OWLIFT_TEMP_MARKER_POINT =
		"owlifttempmarkpoint";
	private static final String PREF_OWLIFT_TEMP_MARKER_X = "owlifttempmarkx";
	private static final String PREF_OWLIFT_TEMP_MARKER_Y = "owlifttempmarky";
	private static final String PREF_OWLIFT_EMISSIVITY = "owliftemissivity";

	// 画面表示
	private boolean useStatusView = true;
	private StatusView statusView;
	private GLView videoView;
	private FrameLayout frameLayout;
	private FrameRate frameRate = new FrameRate();
	private StatusViewUpdater statusViewUpdater;
	private Handler statusViewHandler = new Handler();
	private long tmStart;
	private int frameCnt;
	private boolean running, drawStarted;
	private boolean starting = false;
	private AlertDialogFragment alertDlg;

	// USBホスト
	private static UsbOwliftHost owliftHost;
	private AsyncTask<Integer, Exception, Integer> streamStarter;

	// UVC
	private UvcOwliftFunc.Format format;
	private UsbOwliftDev uvcDev;
	private UvcOwliftFunc uvcFunc;
	private static final boolean decodeStream = true;

	// 端末のマイク
	private boolean useHostMic;
	private OwliftMic hostMic;

	// 録画
	private OwliftMediaRecorder urec;

	// owlift
	private int texWX, texWY;
	private int dispWx;
	private int dispWy;
	private OwliftDialogFragment owliftSettingsDialog;
	private OwliftDisp owliftDisp;
	private OwliftApp.DecodeType decodeType;
	private int decodeRangeMax = OWLIFT_RANGE_MAX_TEMP_DEFAULT;
	private int decodeRangeMin = OWLIFT_RANGE_MIN_TEMP_DEFAULT;

	private boolean isTakePicture = false;

	private int getDispWx()
	{
		return dispWx;
	}

	private int getDispWy()
	{
		return dispWy;
	}

	class StatusView extends View
	{
		private Paint paint = new Paint();
		private StringBuilder line = new StringBuilder(256);

		public StatusView(Context context)
		{
			super(context);
			paint.setColor(Color.WHITE);
			paint.setTextSize(16f);
		}

		@Override
		protected void onDraw(Canvas canvas)
		{
			if (!running)
				return;

			UvcOwliftFunc vf = uvcFunc;

			line.delete(0, line.length());
			if (videoView.getFrameRate() != null) {
				line.append("USB ");
				line.append(frameRate.getFpsString());
				line.append(" FPS / Render ");
				line.append(videoView.getFrameRate().getFpsString());
			} else {
				line.append(frameRate.getFpsString());
			}

			int fc = frameCnt;

			line.append(" FPS / ");
			line.append(" Frame ");
			line.append(fc);
			line.append(" / Time ");
			long t = (System.currentTimeMillis() - tmStart) / 1000;
			line.append(t);

			if (format != null) {
				if (t > 0 && t % 60 == 0) {
					Log.d(
						TAG,
						String.format("%5ds: FRAME=%6d RATE=%5.1f%%",
							t, fc, ((double)fc * format.getInterval())
								/ (t * 100000.0)));
				}
			}

			paint.setTextSize(20);
			canvas.drawText(line.toString(), 5, 20, paint);

			if (vf != null) {
				line.delete(0, line.length());
				if (format != null) {
					line.append(format.getPayloadType());
					line.append(" ");

					line.append(getFormatWidth())
						.append("x")
						.append(getFormatHeight())
						.append(" ")
						.append(String.format(Locale.ENGLISH, "%.1f",
							10000000.0 / format.getInterval()))
						.append(" FPS");
				}

				if (urec != null)
					line.append(" <REC>");
				canvas.drawText(line.toString(), 5, 42, paint);
			}
		}
	}

	private class StreamStarter extends
		AsyncTask<Integer, Exception, Integer>
	{
		private boolean reconfig;

		StreamStarter(boolean reconfig)
		{
			this.reconfig = reconfig;
		}

		@Override
		protected Integer doInBackground(Integer... params)
		{
			try {
				starting = true;
				startDevice(reconfig);
			} catch (Exception e) {
				publishProgress((Exception)e);
				closeDevice();
			} catch (Throwable t) {
				showError(t);
				closeDevice();
			}
			starting = false;
			return 0;
		}

		@Override
		protected void onProgressUpdate(Exception... values)
		{
			showError(values[0]);
		}
	}

	private class StatusViewUpdater extends Thread
	{
		private boolean cancel;

		StatusViewUpdater()
		{
		}

		@Override
		public void run()
		{
			try {
				while (!cancel) {

					UvcOwliftFunc u = uvcFunc;
					if (u != null && u.isStreaming() && drawStarted) {
						statusViewHandler.post(new Runnable() {
							public void run()
							{
								statusView.invalidate();
							}
						});
					}
					try {
						Thread.sleep(1000);
					} catch (InterruptedException e) {
						break;
					}
				}
			} catch (Throwable t) {
				if (needToLog(t))
					Log.d(TAG, "error", t);
			}
		}
	}

	private class FormatInfo
	{
		int menuId;
		UvcOwliftFunc.PayloadType payloadType;
		int wx, wy, intvl;
		UvcOwliftFunc.UncompEncodingType ueType =
			UvcOwliftFunc.UncompEncodingType.UNKNOWN;

		FormatInfo(int menuId, UvcOwliftFunc.PayloadType payloadType, int wx,
			int wy,
			int intvl, UvcOwliftFunc.UncompEncodingType ue)
		{
			this(menuId, payloadType, wx, wy, intvl);
			ueType = ue;
		}

		FormatInfo(int menuId, UvcOwliftFunc.PayloadType payloadType, int wx,
			int wy,
			int intvl)
		{
			this.menuId = menuId;
			this.payloadType = payloadType;
			this.wx = wx;
			this.wy = wy;
			this.intvl = intvl;
		}

		UvcOwliftFunc.Format getFormat()
		{
			if (ueType != UvcOwliftFunc.UncompEncodingType.UNKNOWN) {
				return new UvcOwliftFunc.Format(payloadType, ueType, wx, wy,
					intvl);
			}
			return new UvcOwliftFunc.Format(payloadType, wx, wy, intvl);
		}
	}

	public static class AlertDialogFragment extends DialogFragment
	{
		@Override
		public Dialog onCreateDialog(Bundle savedInstanceState)
		{
			AlertDialog.Builder builder =
				new AlertDialog.Builder(getActivity());

			builder.setTitle(getText(R.string.app_name));

			String version = OwliftApp.getLibVersion();

			builder.setMessage("Version : " + version);
			return builder.create();
		}
	}

	private static String createSaveFileName(int count, String baseName)
	{
		String postfix = "";
		if (count != 0) {
			postfix = "(" + count + ")";
		}
		Calendar calendar = Calendar.getInstance();
		return baseName
			+ String.format(Locale.ENGLISH, "%4d%02d%02d%02d%02d%02d",
				(calendar.get(Calendar.YEAR)),
				(calendar.get(Calendar.MONTH) + 1),
				calendar.get(Calendar.DATE),
				calendar.get(Calendar.HOUR_OF_DAY),
				calendar.get(Calendar.MINUTE),
				calendar.get(Calendar.SECOND))
			+ postfix;
	}

	private FileOutputStream getFileOutputStream(String saveDir)
		throws FileNotFoundException
	{
		FileOutputStream out;
		File dir = new File(saveDir);
		if (!dir.exists()) {
			if (!dir.mkdirs()) {
				showToast(getString(R.string.fail_save_directory),
					Toast.LENGTH_SHORT);
				return null;
			}
		}
		if (!dir.isDirectory()) {
			showToast(getString(R.string.fail_save_directory),
				Toast.LENGTH_SHORT);
			return null;
		}

		String filename = null;
		File saveFile;

		for (int i = 0; i < 1000; i++) {
			filename =
				createSaveFileName(i, "owliftplayer") + ".jpg";
			saveFile = new File(saveDir + "/" + filename);
			if (!saveFile.exists()) {
				break;
			}
			filename = null;
		}

		if (filename == null) {
			showToast(getString(R.string.fail_create_file_for_takepicture),
				Toast.LENGTH_SHORT);
			return null;
		}

		out = new FileOutputStream(new File(dir.getPath(), filename));

		return out;
	}

	private void createPicture(int wx, int wy, byte[] buf)
	{
		if (!isTakePicture) {
			return;
		}
		FileOutputStream out = null;
		boolean finish = false;

		try {
			out = getFileOutputStream(DEFAULT_RECORD_DIR);
			if (out == null) {
				return;
			}

			Bitmap bmp =
				Bitmap.createBitmap(wx, wy,
					android.graphics.Bitmap.Config.RGB_565);
			bmp.copyPixelsFromBuffer(ByteBuffer.wrap(buf));
			bmp.compress(Bitmap.CompressFormat.JPEG, 100, out);
			finish = true;
		} catch (Exception e) {
			showToast(getString(R.string.fail_takepicture), Toast.LENGTH_LONG);
		} finally {
			if (out != null) {
				try {
					out.close();
					if (finish) {
						showToast(
							getString(R.string.finish_takepicture,
								DEFAULT_RECORD_DIR),
							Toast.LENGTH_SHORT);
					}
				} catch (IOException e) {
					showToast(getString(R.string.fail_takepicture),
						Toast.LENGTH_LONG);
				}
			}
			isTakePicture = false;
		}
	}

	private byte[] getDispBuf(byte[] buf)
	{
		byte[] ret = buf;
		if (isOwlift() && buf != null) {
			ret = owliftDisp.getDispBuf(uvcFunc.getDecodeFunc(), ret,
				getFormatWidth(),
				getFormatHeight(), texWX, texWY);
		}
		return ret;
	}

	private class UvcOwliftFuncCallback implements UvcOwliftFunc.Callback
	{
		@Override
		public void recvBuffer(UvcOwliftFunc.Buffer buf)
		{
			frameRate.record();
			++frameCnt;
			if (!running) {
				return;
			}
			if (videoView == null || !videoView.getReady()) {
				return;
			}

			byte[] data = getDispBuf(buf.getData());

			if (data == null) {
				return;
			}
			if (isTakePicture) {
				byte[] pictureBuf = new byte[dispWx * dispWy * 2];
				cpPictureBuf(getDispWx(), getDispWy(), texWX, data,
					pictureBuf);
				createPicture(getDispWx(), getDispWy(),
					pictureBuf);
			} else if (urec != null) {
				int len = data.length;
				if (len > buf.getSize()) {
					// bufのサイズを超えない分だけコピーする
					len = buf.getSize();
				}
				// 拡大した録画データを渡すためにコピー
				// 録画はこの関数が戻った後にライブラリで実行している
				// bufは3倍にしたデータが入るサイズ以上ないといけない。
				// bufはstream開始時に設定している
				System.arraycopy(data, 0, buf.getData(), 0, len);
			}
			videoView.render(data);

			if (!drawStarted)
				drawStarted = true;
		}

		private void cpPictureBuf(int w, int h, int sdw,
			byte[] src, byte[] dst)
		{
			int offset = 0, dstOffset = 0;
			for (int i = 0; i < h; i++) {
				System.arraycopy(src, offset, dst, dstOffset, w * 2);
				offset += sdw * 2;
				dstOffset += w * 2;
			}
		}

		@Override
		public void onException(Throwable t)
		{
			if (needToLog(t))
				showError(t);
		}
	}

	private static boolean needToLog(Throwable t)
	{
		return !(t instanceof OwliftException)
			|| ((OwliftException)t)
				.getExceptionCode() != OwliftExceptionCode.STOPPED;
	}

	private static ViewGroup.LayoutParams createFillLayoutParams()
	{
		return new ViewGroup.LayoutParams(ViewGroup.LayoutParams.MATCH_PARENT,
			ViewGroup.LayoutParams.MATCH_PARENT);
	}

	@Override
	public void onCreate(Bundle savedInstanceState)
	{
		Log.d(TAG, "onCreate");
		super.onCreate(savedInstanceState);

		int count = 0;
		while (isCreated) {
			count++;
			if (count > 300) {
				stopApplication();
				destroyApplication();
				break;
			}
			try {
				Thread.sleep(10);
			} catch (Exception e) {
				// continue;
			}
		}
		isCreated = true;

		setTitle(R.string.title);
		getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

		try {
			owliftHost = UsbOwliftHost.getUsbHost(this);
			owliftHost.onCreate();

		} catch (OwliftException e) {
			showError(e);
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		owliftDisp = new OwliftDisp();
	}

	@Override
	protected void onStart()
	{
		loadPreference();

		Log.d(TAG, "onStart");
		videoView = new GLView(getApplicationContext());
		frameLayout = new FrameLayout(getApplicationContext());
		frameLayout.addView(videoView);
		setContentView(frameLayout);

		if (useStatusView) {
			statusView = new StatusView(getApplicationContext());
			addContentView(statusView, createFillLayoutParams());
		}
		videoView.setOnTouchListener(this);

		streamStarter = new StreamStarter(false).execute();

		super.onStart();
	}

	private DecodeFunc getDecodeFunc()
	{
		if (uvcFunc == null) {
			return null;
		}
		return uvcFunc.getDecodeFunc();
	}

	private void loadPreference()
	{
		SharedPreferences pref = getSharedPreferences(PREF_NAME,
			MODE_PRIVATE);
		useHostMic = pref.getBoolean(PREF_USE_HOSTMIC, false);

		int type = pref.getInt(PREF_OWLIFT_COLOR_TYPE,
			OwliftApp.DecodeType.OWDECODETYPE_BLUE_ORANGE_16.toInt());
		decodeType = OwliftApp.DecodeType.toDecodeType(type);
		int min =
			pref.getInt(PREF_OWLIFT_MIN_TEMP, OWLIFT_RANGE_MIN_TEMP_DEFAULT);
		int max =
			pref.getInt(PREF_OWLIFT_MAX_TEMP, OWLIFT_RANGE_MAX_TEMP_DEFAULT);
		setDecodeRange(min, max);
		setTempUnit(pref.getInt(PREF_OWLIFT_TEMP_UNIT,
				OwliftDisp.TEMP_UNIT_CELSIUS), false);

		setTempMarkerEnable(OwliftDisp.MARKER_TYPE_HIGH,
				pref.getBoolean(PREF_OWLIFT_TEMP_MARKER_HIGH, false), false);
		setTempMarkerEnable(OwliftDisp.MARKER_TYPE_LOW,
				pref.getBoolean(PREF_OWLIFT_TEMP_MARKER_LOW, false), false);
		setDispDate(pref.getBoolean(PREF_OWLFIT_DISPlAY_DATE, false),
				false);
		setTempMarkerEnable(OwliftDisp.MARKER_TYPE_POINT,
				pref.getBoolean(PREF_OWLIFT_TEMP_MARKER_POINT, false), false);

		int x = pref.getInt(PREF_OWLIFT_TEMP_MARKER_X,
			owliftDisp.getOwliftDispWidth() / 2);
		int y = pref.getInt(PREF_OWLIFT_TEMP_MARKER_Y,
			owliftDisp.getOwliftDispHeight() / 2);
		setTempMarker(x, y, false);

		String emissiveStr = pref.getString(PREF_OWLIFT_EMISSIVITY,
				Double.toString(OWLIFT_DEFAULT_EMISSIVITY));
		double emissivity = Double.parseDouble(emissiveStr);
		setOwliftEmissivity(emissivity, false);
	}

	private void savePreference()
	{
		SharedPreferences pref = getSharedPreferences(PREF_NAME,
			MODE_PRIVATE);
		Editor e = pref.edit();
		e.putBoolean(PREF_USE_HOSTMIC, useHostMic);

		e.putInt(PREF_OWLIFT_MIN_TEMP, decodeRangeMin);
		e.putInt(PREF_OWLIFT_MAX_TEMP, decodeRangeMax);
		e.putInt(PREF_OWLIFT_TEMP_UNIT, owliftDisp.getTempUnit());
		if (decodeType != null) {
			e.putInt(PREF_OWLIFT_COLOR_TYPE, decodeType.toInt());
		}
		e.putBoolean(PREF_OWLIFT_TEMP_MARKER_HIGH,
				owliftDisp.getTempMarker(OwliftDisp.MARKER_TYPE_HIGH));
		e.putBoolean(PREF_OWLIFT_TEMP_MARKER_LOW,
				owliftDisp.getTempMarker(OwliftDisp.MARKER_TYPE_LOW));
		e.putBoolean(PREF_OWLIFT_TEMP_MARKER_POINT,
				owliftDisp.getTempMarker(OwliftDisp.MARKER_TYPE_POINT));
		e.putBoolean(PREF_OWLFIT_DISPlAY_DATE, owliftDisp.isDispDate());

		e.apply();
	}

	@Override
	protected void onRestart()
	{
		closeDevice();
		super.onRestart();
	}

	@Override
	protected void onResume()
	{
		super.onResume();
	}

	@Override
	protected void onPause()
	{
		Log.d(TAG, "onPause");
		super.onPause();

		if (starting) {
			int count = 0;
			while (starting) {
				try {
					Thread.sleep(10);
				} catch (Exception e) {
					// nop.
				}
				if (count > 500) {
					break;
				}
				count++;
			}
		}
		dismissDialog(owliftSettingsDialog);
		owliftSettingsDialog = null;

		dismissDialog(alertDlg);
		alertDlg = null;

		isPaused = true;
	}

	private void dismissDialog(DialogFragment dialogFragment)
	{
		if (dialogFragment == null) {
			return;
		}
		Dialog dialog = dialogFragment.getDialog();
		if (dialog == null || !dialog.isShowing()) {
			return;
		}
		dialog.dismiss();
	}

	private void closeDevice()
	{
		running = false;

		if (urec != null) {
			urec.stop();
			urec = null;
		}

		if (hostMic != null) {
			hostMic.close();
			hostMic = null;
		}

		if (owliftHost != null) {
			owliftHost.close();
			uvcDev = null;
			uvcFunc = null;

			owliftDisp.setUsbOwliftDev(null);
		}

	}

	private void stopApplication()
	{
		if (!isPaused) {
			return;
		}
		if (streamStarter != null) {
			streamStarter.cancel(true);
			streamStarter = null;
		}

		if (useStatusView && statusViewUpdater != null) {
			statusViewUpdater.cancel = true;
			statusViewUpdater = null;
		}

		closeDevice();

		if (frameLayout != null) {
			frameLayout.removeView(videoView);
		}
		videoView = null;
	}

	@Override
	protected void onStop()
	{
		Log.d(TAG, "onStop");
		super.onStop();

		stopApplication();
	}

	private void destroyApplication()
	{
		if (!isPaused) {
			return;
		}

		if (owliftHost != null) {
			owliftHost.onDestroy();
			owliftHost = null;
		}

		if (owliftDisp != null) {
			owliftDisp = null;
		}

		isCreated = false;
		isPaused = false;
	}

	@Override
	protected void onDestroy()
	{
		Log.d(TAG, "onDestroy");
		super.onDestroy();
		destroyApplication();
	}

	private void showError(Throwable e)
	{
		String s = e.getMessage();

		Log.e(TAG, "error", e);

		if (s == null || s.length() == 0)
			s = "error: " + e.getClass().getName();

		final Context ctxt = this;
		final String msg = s;

		statusViewHandler.post(new Runnable() {
			@Override
			public void run() {
				if (videoView == null) {
					return;
				}
				AlertDialog.Builder a = new AlertDialog.Builder(ctxt);
				a.setTitle(getText(R.string.app_name)).setMessage(msg).show();
			}
		});
	}

	private void showToast(String s, int length)
	{
		final Context ctxt = this;
		final String msg = s;
		final int len = length;

		statusViewHandler.post(new Runnable() {
			@Override
			public void run() {
				int x;
				switch (len) {
					case Toast.LENGTH_SHORT:
						x = Toast.LENGTH_SHORT;
						break;
					case Toast.LENGTH_LONG:
					default:
						x = Toast.LENGTH_LONG;
						break;
				}
				Toast.makeText(ctxt, msg, x).show();
			}
		});
	}

	private void openUvcDevice(UsbOwliftHost uh,
		long timeoutMSecWaitUntilDeviceFound)
			throws OwliftException
	{
		uvcDev = uh.openDev(this, timeoutMSecWaitUntilDeviceFound,
			new UsbOwliftHost.Callback() {
				@Override
				public UsbOwliftDev selectDev(UsbOwliftDev.Collection devs)
				{
					UsbOwliftDev.Collection vdevs =
						devs.getDev();
					if (vdevs.isEmpty())
						return null;
					return vdevs.get(0);
				}
			});
		owliftDisp.setUsbOwliftDev(uvcDev);

		if (isFinishing())
			return;

		for (UsbOwliftFunc func : uvcDev.getFuncs()) {
			if (func != null) {
				uvcFunc = (UvcOwliftFunc)func;
				int vid = getVID();
				int pid = getPID();
				if (vid == OwliftDisp.OWLIFT_VID && pid == OwliftDisp.OWLIFT_PID_B) {
					if (getDecodeFunc() != null) {
						getDecodeFunc().setCircumferenceCorrection(false);
					}
				}
				break;
			}
		}

		if (uvcFunc == null)
			throw new IllegalStateException("No UVC device");

	}

	private int getPID()
	{
		if (uvcDev == null) {
			return -1;
		}
		return uvcDev.getPID();
	}

	private int getVID()
	{
		if (uvcDev == null) {
			return -1;
		}
		return uvcDev.getVID();
	}

	private void setDecodeRange(int min, int max)
	{
		if (getDecodeFunc() == null) {
			return;
		}

		getDecodeFunc().setDecodeRange(min, max);
	}

	private int getDecodeRangeMin()
	{
		return decodeRangeMin;
	}

	private void setDecodeRangeMin(int gainControl)
	{
		if (getDecodeFunc() == null || gainControl == decodeRangeMin) {
			return;
		}
		if (gainControl >= this.decodeRangeMax) {
			gainControl = this.decodeRangeMax - 1;
		}
		getDecodeFunc().setDecodeRange(gainControl, decodeRangeMax);
		decodeRangeMin = gainControl;
		savePreference();
	}

	private int getDecodeRangeMax()
	{
		return decodeRangeMax;
	}

	private void setManualGainMax(int gainControl)
	{
		if (getDecodeFunc() == null || gainControl == decodeRangeMax) {
			return;
		}
		if (gainControl <= this.decodeRangeMin) {
			gainControl = this.decodeRangeMin + 1;
		}
		getDecodeFunc().setDecodeRange(decodeRangeMin, gainControl);
		decodeRangeMax = gainControl;
		savePreference();
	}

	private int getTempUnit()
	{
		return owliftDisp.getTempUnit();
	}

	private void setTempUnit(int tempUnit)
	{
		setTempUnit(tempUnit, true);
	}

	private void setTempUnit(int tempUnit, boolean isSave)
	{
		owliftDisp.setTempUnit(tempUnit);
		if (isSave)
			savePreference();
	}

	OwliftApp.DecodeType getDecodeType()
	{
		return decodeType;
	}

	private void setDecodeType(OwliftApp.DecodeType decodeType)
	{
		if (decodeType == this.decodeType) {
			return;
		}

		if (getDecodeFunc() != null) {
			getDecodeFunc().setDecodeType(decodeType);
		}
		this.decodeType = decodeType;
		savePreference();
	}

	private boolean getTempMarker(int type)
	{
		return owliftDisp.getTempMarker(type);
	}

	private void setTempMarkerEnable(int type, boolean enable)
	{
		setTempMarkerEnable(type, enable, true);
	}

	private void setTempMarkerEnable(int type, boolean enable,
		boolean isSave)
	{
		owliftDisp.setTempMarkerEnable(type, enable);
		if (isSave)
			savePreference();
	}

	private boolean isDispDate()
	{
		return owliftDisp.isDispDate();
	}

	private void setDispDate(boolean isDisp)
	{
		setDispDate(isDisp, true);
	}

	private void setDispDate(boolean isDisp, boolean isSave)
	{
		owliftDisp.setDispDate(isDisp);
		if (isSave)
			savePreference();
	}

	private void setTempMarker(int x, int y)
	{
		setTempMarker(x, y, true);
	}

	private void setTempMarker(int x, int y, boolean isSave)
	{
		if (x < 0 || y < 0 ||
			x > owliftDisp.getOwliftDispWidth() ||
			y > owliftDisp.getOwliftDispHeight()) {
			return;
		}
		owliftDisp.setTempMarker(x, y);
		if (isSave) {
			SharedPreferences pref = getSharedPreferences(PREF_NAME,
				Context.MODE_PRIVATE);
			Editor e = pref.edit();
			e.putInt(PREF_OWLIFT_TEMP_MARKER_X, x);
			e.putInt(PREF_OWLIFT_TEMP_MARKER_Y, y);

			e.apply();
		}
	}

	double getEmissivity()
	{
		return owliftDisp.getEmissivity();
	}

	void setOwliftEmissivity(double value)
	{
		setOwliftEmissivity(value, true);
	}

	private void setOwliftEmissivity(double value, boolean isSave)
	{
		if (value > 1 || value <= 0) {
			return;
		}
		owliftDisp.setEmissivity(value);
		if (isSave) {
			SharedPreferences pref =
					getSharedPreferences(PREF_NAME, Context.MODE_PRIVATE);
			Editor e = pref.edit();
			e.putString(PREF_OWLIFT_EMISSIVITY, Double.toString(value));

			e.apply();
		}
	}

	private static boolean isOwliftPid(int pid)
	{
		return (pid == OwliftDisp.OWLIFT_PID
				|| pid == OwliftDisp.OWLIFT_PID_B
				|| pid == OwliftDisp.OWLIFT_PID_E
				|| pid == OwliftDisp.OWLIFT_PID_D
				|| pid == OwliftDisp.OWLIFT_PID_F
		);
	}

	private boolean isOwlift()
	{
		return format != null
			&& UncompEncodingType.USER_UE == format.getUncompEncodingType()
			&& getVID() == OwliftDisp.OWLIFT_VID &&
			isOwliftPid(getPID());
	}

	public int getFormatWidth()
	{
		if (format == null) {
			return 0;
		}
		return format.getWidth();
	}

	public int getFormatHeight()
	{
		if (format == null) {
			return 0;
		}
		int h = format.getHeight();
		if (h == owliftDisp.getOwliftRawHeight()) {
			h = owliftDisp.getOwliftDispHeight();
		}
		return h;
	}

	private int normalizeTextureSize(int x)
	{
		if (x < 256)
			return 256;
		else if (x < 512)
			return 512;
		else if (x < 1024)
			return 1024;
		return 2048;
	}

	private void startDevice(boolean reconfig) throws OwliftException
	{
		if (reconfig) {
			closeDevice();
			frameRate.reset();
			frameCnt = 0;
			owliftHost = UsbOwliftHost.getUsbHost(this);
		}

		long timeoutMSecWaitUntilDeviceFound = reconfig ? 10000 : 0;

		openUvcDevice(owliftHost, timeoutMSecWaitUntilDeviceFound);

		if (useHostMic) {
			hostMic = new OwliftMic();
			hostMic.startInStream(44100, 2, 2, new OwliftMic.Callback() {

				@Override
				public void recvBuffer(OwliftMic.Buffer buf)
				{
				}

				@Override
				public void onException(Throwable t)
				{
					showError(t);
				}
			});
		}

		setFormat();
		dispWx = getFormatWidth();
		dispWy = getFormatHeight();
		if (isOwlift()) {
			dispWx = getFormatWidth() * OWLIFT_MAGNIFY_SIZE;
			dispWy = dispWy * OWLIFT_MAGNIFY_SIZE;
		}
		texWX = normalizeTextureSize(dispWx);
		texWY = normalizeTextureSize(dispWy);

		videoView.setImageSize(dispWx, dispWy, texWX, texWY);
		UvcOwliftFunc.UncompEncodingType t = format.getUncompEncodingType();
		if (UvcOwliftFunc.UncompEncodingType.USER_UE == t && isOwlift()) {
			// 表示画面を録画するための関数
			uvcFunc.setPostRecFrame(dispWx, dispWy, texWX);
			uvcFunc.setDecodeType(decodeType);

			uvcFunc.startStream(format,
				texWX, texWY,
				decodeStream, new UvcOwliftFuncCallback());
		} else {
			// OWLIFTのみ対応
			throw new IllegalStateException();
		}

		tmStart = System.currentTimeMillis();
		running = true;

		if (useStatusView && statusViewUpdater == null) {
			statusViewUpdater = new StatusViewUpdater();
			statusViewUpdater.start();
		}

	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu)
	{
		MenuInflater inflater = getMenuInflater();
		inflater.inflate(R.menu.menu_main, menu);

		Thread.setDefaultUncaughtExceptionHandler(
				new Thread.UncaughtExceptionHandler() {

					@Override
					public void uncaughtException(Thread thread, Throwable ex) {
						Log.d(TAG, "error", ex);
					}
				});

		return true;
	}

	@Override
	public boolean onPrepareOptionsMenu(Menu menu)
	{
		menu.findItem(R.id.menu_recstart)
			.setVisible(drawStarted && (urec == null));
		menu.findItem(R.id.menu_recstop).setVisible(
				drawStarted && (urec != null) && urec != null);
		menu.findItem(R.id.menu_takepicture).setVisible(drawStarted
				&& (uvcFunc != null) && (urec == null));
		if (useHostMic) {
			menu.findItem(R.id.menu_use_internal_mic)
				.setTitle(getString(R.string.menu_unuse_internal_mic));
		} else {
			menu.findItem(R.id.menu_use_internal_mic)
				.setTitle(getString(R.string.menu_use_internal_mic));
		}
		menu.findItem(R.id.menu_readffcmode).setVisible(drawStarted && (uvcFunc != null) &&
				(urec == null));
		menu.findItem(R.id.menu_runffc).setVisible(drawStarted && (uvcFunc != null) &&
				(urec == null));

		return true;
	}

	private void startRecording() throws OwliftException
	{
		try {
			urec = new OwliftMediaRecorder(uvcFunc, hostMic);

			File dir = new File(DEFAULT_RECORD_DIR);
			boolean ret = true;
			if (!dir.exists()) {
				ret = dir.mkdirs();// 失敗したら次のチェックでエラーになる
			}
			if (!ret) {
				showToast("Can not create directory.", Toast.LENGTH_SHORT);
				return;
			}
			urec.openFile(DEFAULT_RECORD_DIR
					+ "/" + createSaveFileName(0, "owliftplayer") + ".mp4");
			urec.setVideoFrameRate(10);
			// 変換後のデータを録画
			urec.setVideoPostRecvBuffer(true);

			urec.start();
			frameCnt = 0;
			tmStart = System.currentTimeMillis();
		} catch (OwliftException e) {
			if (urec != null) {
				urec.stop();
				urec = null;
			}
			throw e;
		}
	}

	private void stopRecording()
	{
		if (urec != null) {
			urec.stop();
			urec = null;
			showToast(getString(R.string.stop_recording, DEFAULT_RECORD_DIR),
				Toast.LENGTH_SHORT);
		}
	}

	private void setFormat()
	{
		int owliftWidth = owliftDisp.getOwliftRawWidth();
		int owliftHeight = owliftDisp.getOwliftRawHeight();
		final FormatInfo[] fmtInfos =
			{
				new FormatInfo(R.id.menu_owlift_uncomp_80x60,
					UvcOwliftFunc.PayloadType.UNCOMPRESSED,
					owliftWidth, owliftHeight,
					1158301, UncompEncodingType.USER_UE),
			};

		FormatInfo fmtInfo = null;

		int formatId = R.id.menu_owlift_uncomp_80x60;
		for (FormatInfo fmtInfo0 : fmtInfos) {
			if (fmtInfo0.menuId == formatId) {
				fmtInfo = fmtInfo0;
				break;
			}
		}

		if (fmtInfo != null)
			format = fmtInfo.getFormat();
	}

	private void takePicture()
	{
		showToast("Taking a picture ...", Toast.LENGTH_LONG);
		if (!isTakePicture) {
			isTakePicture = true;
		}
	}

	private void showAboutDialog()
	{
		alertDlg = new AlertDialogFragment();
		alertDlg.show(getFragmentManager(), "about");

	}

	private boolean onOptionsItemSelected0(MenuItem item)
		throws OwliftException
	{
		switch (item.getItemId()) {
		case R.id.menu_recstart:
			startRecording();
			break;

		case R.id.menu_recstop:
			stopRecording();
			break;

		case R.id.menu_takepicture:
			takePicture();
			break;

		case R.id.menu_use_internal_mic:
			useHostMic = !useHostMic;
			savePreference();
			break;

		case R.id.menu_owliftsettings:
			showOwliftSettingDialog();
			break;

		case R.id.menu_about:
			showAboutDialog();
			break;

		case R.id.menu_readffcmode:
			readFFCMode();
			break;
		case R.id.menu_runffc:
			runFFC();
			break;
		default:
			return false;
		}

		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item)
	{
		try {
			return onOptionsItemSelected0(item)
				|| super.onOptionsItemSelected(item);
		} catch (Exception e) {
			showError(e);
		}
		return false;
	}

	@Override
	public void onDialogClicked(String tag, Bundle args, int which)
	{

	}

	@Override
	public void onDialogCancelled(String tag, Bundle args)
	{

	}

	@Override
	public void onCreateDialogView(View view, int viewId)
	{
		if (R.layout.owlift == viewId && owliftSettingsDialog != null) {
			setupOwliftSettings(view);
		}
	}

	@Override
	public void onActivityCreated(View view, int viewId)
	{
		if (R.layout.owlift == viewId) {
			if (owliftSettingsDialog != null) {
				Window w = owliftSettingsDialog.getDialog().getWindow();
				if (w == null)
					return;
				WindowManager.LayoutParams lp = w.getAttributes();
				lp.gravity = Gravity.START;
				w.setAttributes(lp);
			}
		}
	}

	@Override
	public void onStart(Dialog dialog, Bundle args)
	{

	}

	@Override
	public void onDialogNeutralClicked(String tag, Bundle args, int which)
	{

	}

	private int getGainControlMin()
	{
		return getDecodeRangeMin() - OWLIFT_TEMP_MIN;
	}

	private int getGainControlMax()
	{
		return getDecodeRangeMax() - OWLIFT_TEMP_MIN;
	}

	private double convK100(int k100, int tempUnit)
	{
		double ret = 0;
		if (k100 > 200) {
			if (tempUnit == OwliftDisp.TEMP_UNIT_CELSIUS) {
				ret = k100 / 100.0 - 273.15;
			} else if (tempUnit == OwliftDisp.TEMP_UNIT_FAHLENHEIT) {
				ret = (k100 / 100.0 - 273.15) * 1.8 + 32;
			} else {
				ret = (double)k100 / 100;
			}
		}
		return ret;
	}

	private String convK100String(int k100, int tempUnit)
	{
		String tempUnitType = "C";
		if (tempUnit == OwliftDisp.TEMP_UNIT_FAHLENHEIT) {
			tempUnitType = "F";
		} else if (tempUnit == OwliftDisp.TEMP_UNIT_KELVIN) {
			tempUnitType = "K";
		}
		BigDecimal bd = new BigDecimal(convK100(k100, tempUnit));
		return bd.setScale(1, BigDecimal.ROUND_HALF_UP).toString() +
			tempUnitType;
	}

	private void showOwliftSettingDialog()
	{
		OwliftDialogFragment.Builder builder =
			new OwliftDialogFragment.Builder();
		builder.setViewId(R.layout.owlift);
		builder.setPositiveButton(true);
		builder.setNegativeButton(false);
		builder.setNeutralButton(false);
		owliftSettingsDialog = builder.create();
		owliftSettingsDialog.show(getFragmentManager(), "owlift");
	}

	private void setDecodeSettingsVisible(View view, boolean isAgcChecked)
	{
		TextView min = (TextView)view.findViewById(R.id.textView_min);
		TextView max = (TextView)view.findViewById(R.id.textView_max);
		TextView minRange =
			(TextView)view.findViewById(R.id.textView_min_range);
		TextView maxRange =
			(TextView)view.findViewById(R.id.textView_max_range);
		SeekBar inputMin = (SeekBar)view.findViewById(R.id.seekBar_input_min);
		SeekBar inputMax = (SeekBar)view.findViewById(R.id.seekBar_input_max);

		TextView rangeText =
			(TextView)view.findViewById(R.id.textView_InputRange);
		LinearLayout owliftRangeMin = (LinearLayout)view.findViewById(
			R.id.linearLayout_owliftRangeMin);
		LinearLayout owliftRangeMax = (LinearLayout)view.findViewById(
			R.id.linearLayout_owliftRangeMax);
		RadioGroup agcGroup =
			(RadioGroup)view.findViewById(R.id.radioGroup_decodeType);
		if (isAgcChecked) {
			owliftRangeMin.setVisibility(View.GONE);
			owliftRangeMax.setVisibility(View.GONE);
			rangeText.setVisibility(View.GONE);
			agcGroup.setVisibility(View.VISIBLE);

			min.setEnabled(false);
			max.setEnabled(false);
			minRange.setEnabled(false);
			maxRange.setEnabled(false);
			inputMin.setEnabled(false);
			inputMax.setEnabled(false);
		} else {
			owliftRangeMin.setVisibility(View.VISIBLE);
			owliftRangeMax.setVisibility(View.VISIBLE);
			rangeText.setVisibility(View.VISIBLE);
			agcGroup.setVisibility(View.GONE);

			min.setEnabled(true);
			max.setEnabled(true);
			minRange.setEnabled(true);
			maxRange.setEnabled(true);
			inputMin.setEnabled(true);
			inputMax.setEnabled(true);
		}
	}

	private OwliftApp.DecodeType getDecodeTypeFromSettingDialog(boolean isAgc,
		int colorTypeId, int decodeTypeId)
	{
		OwliftApp.DecodeType decodeType =
			OwliftApp.DecodeType.OWDECODETYPE_NON_LINEAR_AGC_BLUE_ORANGE_16;
		boolean linearAGC = (decodeTypeId == R.id.radio_decodeType_LAGC);
		if (colorTypeId == R.id.radio_colorType_blue_orange) {
			decodeType = isAgc
				? linearAGC
					? OwliftApp.DecodeType.OWDECODETYPE_LINEAR_AGC_BLUE_ORANGE_16
					: OwliftApp.DecodeType.OWDECODETYPE_NON_LINEAR_AGC_BLUE_ORANGE_16
				: OwliftApp.DecodeType.OWDECODETYPE_BLUE_ORANGE_16;
		} else if (colorTypeId == R.id.radio_colorType_rainbow) {
			decodeType = isAgc
				? linearAGC
					? OwliftApp.DecodeType.OWDECODETYPE_LINEAR_AGC_RAINBOW_16
					: OwliftApp.DecodeType.OWDECODETYPE_NON_LINEAR_AGC_RAINBOW_16
				: OwliftApp.DecodeType.OWDECODETYPE_RAINBOW_16;
		} else if (colorTypeId == R.id.radio_colorType_gray) {
			decodeType = isAgc
				? linearAGC
					? OwliftApp.DecodeType.OWDECODETYPE_LINEAR_AGC_GRAY_16
					: OwliftApp.DecodeType.OWDECODETYPE_NON_LINEAR_AGC_GRAY_16
				: OwliftApp.DecodeType.OWDECODETYPE_GRAY_16;
		}
		return decodeType;
	}

	// 放射率の小数点が設定された場合に整数値を0にする。
	private void changeEmissivityInteger(int oldVal, int newVal)
	{
		if (oldVal == 0 && newVal != 0 && owliftSettingsDialog != null) {
			Dialog dlg = owliftSettingsDialog.getDialog();
			NumberPicker npi =
					(NumberPicker)dlg.findViewById(R.id.numberPicker_integer);
			if (npi.getValue() == npi.getMaxValue()) {
				npi.setValue(0);
			}
		}
	}

	private void setEmissivity(OwliftDialogFragment dialog)
	{
		if (dialog == null) {
			return;
		}
		Dialog dlg = dialog.getDialog();
		NumberPicker npi =
				(NumberPicker)dlg.findViewById(R.id.numberPicker_integer);
		NumberPicker npo =
				(NumberPicker)dlg.findViewById(R.id.numberPicker_one);
		NumberPicker npt =
				(NumberPicker)dlg.findViewById(R.id.numberPicker_two);
		BigDecimal emi1 = new BigDecimal(Double.toString(npi.getValue())).multiply(new BigDecimal(100));
		BigDecimal emi2 = new BigDecimal(Double.toString(npo.getValue())).multiply(new BigDecimal(10));
		BigDecimal emi3 = new BigDecimal(Double.toString(npt.getValue()));
		BigDecimal calc = emi1.add(emi2).add(emi3).divide(new BigDecimal(100));
		double emissivity = calc.doubleValue();
		if (emissivity > 1 || emissivity <= 0) {
			return;
		}
		setOwliftEmissivity(emissivity);
	}

	private void setupOwliftSettings(View view)
	{
		final TextView minRange = (TextView)view.findViewById(
			R.id.textView_min_range);
		minRange.setText(convK100String(getDecodeRangeMin(),
			getTempUnit()));
		final TextView maxRange = (TextView)view.findViewById(
			R.id.textView_max_range);
		maxRange.setText(convK100String(getDecodeRangeMax(),
			getTempUnit()));

		RadioGroup colorType =
			(RadioGroup)view.findViewById(R.id.radioGroup_colorType);
		RadioGroup decodeGroup =
			(RadioGroup)view.findViewById(R.id.radioGroup_decodeType);
		decodeGroup.check(R.id.radio_decodeType_LAGC);
		CheckBox checkAGC = (CheckBox)view.findViewById(R.id.checkBox_AGC);
		checkAGC.setChecked(false);
		switch (getDecodeType()) {
		case OWDECODETYPE_NON_LINEAR_AGC_BLUE_ORANGE_16:
			decodeGroup.check(R.id.radio_decodeType_NLAGC);
		case OWDECODETYPE_LINEAR_AGC_BLUE_ORANGE_16:
			checkAGC.setChecked(true);
		case OWDECODETYPE_BLUE_ORANGE_16:
			colorType.check(R.id.radio_colorType_blue_orange);
			break;
		case OWDECODETYPE_NON_LINEAR_AGC_RAINBOW_16:
			decodeGroup.check(R.id.radio_decodeType_NLAGC);
		case OWDECODETYPE_LINEAR_AGC_RAINBOW_16:
			checkAGC.setChecked(true);
		case OWDECODETYPE_RAINBOW_16:
			colorType.check(R.id.radio_colorType_rainbow);
			break;
		case OWDECODETYPE_NON_LINEAR_AGC_GRAY_16:
			decodeGroup.check(R.id.radio_decodeType_NLAGC);
		case OWDECODETYPE_LINEAR_AGC_GRAY_16:
			checkAGC.setChecked(true);
		case OWDECODETYPE_GRAY_16:
			colorType.check(R.id.radio_colorType_gray);
			break;
		default:
			break;
		}
		setDecodeSettingsVisible(view, checkAGC.isChecked());

		colorType.setOnCheckedChangeListener(
			new RadioGroup.OnCheckedChangeListener() {

				@Override
				public void onCheckedChanged(RadioGroup group, int checkedId)
				{
					OwliftApp.DecodeType decodeType;
					Dialog owliftSettings = owliftSettingsDialog.getDialog();
					CheckBox agc = (CheckBox)owliftSettings.findViewById(
						R.id.checkBox_AGC);
					RadioGroup groupDecodeType =
						(RadioGroup)owliftSettings
							.findViewById(R.id.radioGroup_decodeType);
					int checkedAGCId =
						groupDecodeType.getCheckedRadioButtonId();
					boolean isAgc = agc.isChecked();
					decodeType = getDecodeTypeFromSettingDialog(isAgc,
						checkedId, checkedAGCId);
					setDecodeType(decodeType);
				}
			});

		decodeGroup.setOnCheckedChangeListener(
			new RadioGroup.OnCheckedChangeListener() {
				@Override
				public void onCheckedChanged(RadioGroup group, int checkedId)
				{
					OwliftApp.DecodeType decodeType;
					Dialog owliftSettings = owliftSettingsDialog.getDialog();
					CheckBox agc = (CheckBox)owliftSettings.findViewById(
						R.id.checkBox_AGC);
					RadioGroup colorType =
						(RadioGroup)owliftSettings
							.findViewById(R.id.radioGroup_colorType);
					int checkedColorId = colorType.getCheckedRadioButtonId();
					boolean isAgc = agc.isChecked();
					if (!isAgc) {
						return;
					}
					decodeType = getDecodeTypeFromSettingDialog(true,
						checkedColorId, checkedId);
					setDecodeType(decodeType);
				}
			});

		checkAGC.setOnCheckedChangeListener(
			new CompoundButton.OnCheckedChangeListener() {

				@Override
				public void onCheckedChanged(CompoundButton buttonView,
					boolean isChecked)
				{
					Dialog owliftSettings = owliftSettingsDialog.getDialog();
					RadioGroup colorType =
						(RadioGroup)owliftSettings
							.findViewById(R.id.radioGroup_colorType);
					int checkedId = colorType.getCheckedRadioButtonId();
					RadioGroup groupDecodeType =
						(RadioGroup)owliftSettings
							.findViewById(R.id.radioGroup_decodeType);
					int checkedAGCId =
						groupDecodeType.getCheckedRadioButtonId();
					OwliftApp.DecodeType decodeType =
						getDecodeTypeFromSettingDialog(isChecked, checkedId,
							checkedAGCId);
					setDecodeType(decodeType);
					setDecodeSettingsVisible(
						owliftSettings.findViewById(R.id.owlift_scrollview),
						isChecked);
				}
			});

		RadioGroup tempUnit =
			(RadioGroup)view.findViewById(R.id.radioGroup_tempUnit);
		switch (getTempUnit()) {
		case OwliftDisp.TEMP_UNIT_CELSIUS:
			tempUnit.check(R.id.radio_tempUnit_celsius);
			break;
		case OwliftDisp.TEMP_UNIT_FAHLENHEIT:
			tempUnit.check(R.id.radio_tempUnit_fahrenheit);
			break;
		case OwliftDisp.TEMP_UNIT_KELVIN:
			tempUnit.check(R.id.radio_tempUnit_kelvin);
			break;
		}
		tempUnit.setOnCheckedChangeListener(
			new RadioGroup.OnCheckedChangeListener() {
				@Override
				public void onCheckedChanged(RadioGroup group, int checkedId)
				{
					int tempUnit = OwliftDisp.TEMP_UNIT_CELSIUS;
					if (checkedId == R.id.radio_tempUnit_celsius) {
						tempUnit = OwliftDisp.TEMP_UNIT_CELSIUS;
					} else if (checkedId == R.id.radio_tempUnit_fahrenheit) {
						tempUnit = OwliftDisp.TEMP_UNIT_FAHLENHEIT;
					} else if (checkedId == R.id.radio_tempUnit_kelvin) {
						tempUnit = OwliftDisp.TEMP_UNIT_KELVIN;
					}
					if (tempUnit != getTempUnit()) {
						minRange.setText(convK100String(getDecodeRangeMin(),
							tempUnit));
						maxRange.setText(convK100String(getDecodeRangeMax(),
							tempUnit));
					}
					setTempUnit(tempUnit);
				}
			});

		SeekBar inputMin = (SeekBar)view.findViewById(R.id.seekBar_input_min);
		inputMin.setMax(OWLIFT_TEMP_MAX - OWLIFT_TEMP_MIN);
		inputMin.setProgress(getGainControlMin());
		inputMin
			.setOnSeekBarChangeListener(new SeekBar.OnSeekBarChangeListener() {

				@Override
				public void onStopTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onStartTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onProgressChanged(SeekBar seekBar, int progress,
					boolean fromUser)
				{
					int tempMin = progress + OWLIFT_TEMP_MIN;
					int tempMax = getDecodeRangeMax();
					if (tempMax < tempMin) {
						tempMin = tempMax;
						seekBar.setProgress(tempMin - OWLIFT_TEMP_MIN);
					}
					setDecodeRangeMin(tempMin);
					minRange.setText(convK100String(getDecodeRangeMin(),
						getTempUnit()));
				}
			});

		SeekBar inputMax = (SeekBar)view.findViewById(R.id.seekBar_input_max);
		inputMax.setMax(OWLIFT_TEMP_MAX - OWLIFT_TEMP_MIN);
		inputMax.setProgress(getGainControlMax());
		inputMax
			.setOnSeekBarChangeListener(new SeekBar.OnSeekBarChangeListener() {

				@Override
				public void onStopTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onStartTrackingTouch(SeekBar seekBar)
				{
				}

				@Override
				public void onProgressChanged(SeekBar seekBar, int progress,
					boolean fromUser)
				{
					int tempMax = progress + OWLIFT_TEMP_MIN;
					int tempMin = getDecodeRangeMin();
					if (tempMin > tempMax) {
						tempMax = tempMin;
						seekBar.setProgress(tempMax - OWLIFT_TEMP_MIN);
					}
					setManualGainMax(tempMax);
					maxRange.setText(convK100String(getDecodeRangeMax(),
						getTempUnit()));
				}
			});

		Switch highSwitch =
			(Switch)view.findViewById(R.id.switch_temp_indicator_high);
		highSwitch.setChecked(getTempMarker(OwliftDisp.MARKER_TYPE_HIGH));
		highSwitch.setOnCheckedChangeListener(
			new CompoundButton.OnCheckedChangeListener() {

				@Override
				public void onCheckedChanged(CompoundButton buttonView,
					boolean isChecked)
				{
					setTempMarkerEnable(OwliftDisp.MARKER_TYPE_HIGH,
						isChecked);
				}
			});
		Switch lowSwitch =
			(Switch)view.findViewById(R.id.switch_temp_indicator_low);
		lowSwitch.setChecked(getTempMarker(OwliftDisp.MARKER_TYPE_LOW));
		lowSwitch.setOnCheckedChangeListener(
			new CompoundButton.OnCheckedChangeListener() {

				@Override
				public void onCheckedChanged(CompoundButton buttonView,
					boolean isChecked)
				{
					setTempMarkerEnable(OwliftDisp.MARKER_TYPE_LOW,
						isChecked);
				}
			});
		Switch pointSwitch =
			(Switch)view.findViewById(R.id.switch_temp_indicator_point);
		pointSwitch
			.setChecked(getTempMarker(OwliftDisp.MARKER_TYPE_POINT));
		pointSwitch.setOnCheckedChangeListener(
			new CompoundButton.OnCheckedChangeListener() {

				@Override
				public void onCheckedChanged(CompoundButton buttonView,
					boolean isChecked)
				{
					setTempMarkerEnable(OwliftDisp.MARKER_TYPE_POINT,
						isChecked);
				}
			});

		Switch dispDate = (Switch)view.findViewById(R.id.switch_display_date);
		dispDate.setChecked(isDispDate());
		dispDate.setOnCheckedChangeListener(
			new CompoundButton.OnCheckedChangeListener() {

				@Override
				public void onCheckedChanged(CompoundButton buttonView,
					boolean isChecked)
				{
					setDispDate(isChecked);
				}
			});

		NumberPicker npi =
				(NumberPicker)view.findViewById(R.id.numberPicker_integer);
		NumberPicker npo =
				(NumberPicker)view.findViewById(R.id.numberPicker_one);
		NumberPicker npt =
				(NumberPicker)view.findViewById(R.id.numberPicker_two);

		npi.setMaxValue(1);
		npi.setMinValue(0);
		npo.setMaxValue(9);
		npo.setMinValue(0);
		npt.setMaxValue(9);
		npt.setMinValue(0);
		double emissivity = getEmissivity();
		npi.setValue((int)emissivity);
		npo.setValue(((int)(emissivity * 10) % 10));
		npt.setValue((int)(emissivity * 100) % 10);


		npi.setOnValueChangedListener(
				new NumberPicker.OnValueChangeListener() {
					@Override
					public void onValueChange(NumberPicker picker, int oldVal, int newVal)
					{
						if (newVal == picker.getMaxValue()) {
							Dialog dlg = owliftSettingsDialog.getDialog();
							NumberPicker npo =
									(NumberPicker)dlg.findViewById(R.id.numberPicker_one);
							NumberPicker npt =
									(NumberPicker)dlg.findViewById(R.id.numberPicker_two);
							npo.setValue(0);
							npt.setValue(0);
							setEmissivity(owliftSettingsDialog);
						}
					}
				});

		npo.setOnValueChangedListener(
				new NumberPicker.OnValueChangeListener() {

					@Override
					public void onValueChange(NumberPicker picker, int oldVal, int newVal)
					{
						changeEmissivityInteger(oldVal, newVal);
						setEmissivity(owliftSettingsDialog);
					}
				});

		npt.setOnValueChangedListener(
				new NumberPicker.OnValueChangeListener() {

					@Override
					public void onValueChange(NumberPicker picker, int oldVal, int newVal)
					{
						changeEmissivityInteger(oldVal, newVal);
						setEmissivity(owliftSettingsDialog);
					}
				});

	}

	@Override
	public boolean onTouch(View v, MotionEvent event)
	{
		super.onTouchEvent(event);

		if (isOwlift()) {
			if (getTempMarker(OwliftDisp.MARKER_TYPE_POINT) &&
				(event.getAction()
					& MotionEvent.ACTION_MASK) == MotionEvent.ACTION_DOWN) {
				int[] coordinates = new int[2];
				videoView.convertFrameCoordinates(event.getX(), event.getY(),
					getFormatWidth(), getFormatHeight(), coordinates);
				setTempMarker(coordinates[0], coordinates[1]);
			}
		}
		return false;
	}

	public void readFFCMode()
	{
		int sz = 16;
		int[] data = new int[sz];

		int ret = uvcDev.commandGet(0x0200, 0x3C, data, sz);
		if (ret != OwliftErrorId.ERR_OK) {
			showToast("Error read FFC Mode " + ret,
					Toast.LENGTH_SHORT);
			return;
		}
		StringBuilder sb = new StringBuilder();
		sb.append("Read FFC Mode\n");
		for (int i = 0; i < sz; i++) {
			if (i != 0) {
				sb.append(" ");
			}
			sb.append(String.format("%04x", data[i]));
		}

		showToast(sb.toString(), Toast.LENGTH_SHORT);
	}

	public void runFFC()
	{
		int ret = uvcDev.commandRun(0x0200, 0x40);
		if (ret != OwliftErrorId.ERR_OK) {
			showToast("Error run FFC " + ret, Toast.LENGTH_SHORT);
		}
	}

}
